#include <v8.h>
#include <node.h>
#include <node_events.h>

#include <string>

#include <unistd.h>

#include <vlc/vlc.h>

using namespace node;
using namespace v8;
using namespace std;

static Persistent<String> data_symbol;
static Persistent<String> play_symbol;
static Persistent<String> add_symbol;
static Persistent<String> stop_symbol;
static Persistent<String> end_symbol;

#define REQ_FUN_ARG(I, VAR)                                             \
  if (args.Length() <= (I) || !args[I]->IsFunction())                   \
    return ThrowException(Exception::TypeError(                         \
                  String::New("Argument " #I " must be a function")));  \
  Local<Function> VAR = Local<Function>::Cast(args[I]);

class NodeVLC: public EventEmitter
{
	private:
		libvlc_instance_t *inst;
		libvlc_media_list_player_t *media_list_player;
		libvlc_media_list_t *media_list;
		libvlc_media_player_t *media_player;
		ev_timer poll_timer;
		int poll_rate;
	public:

		static void Init(Handle<Object> target)
		{
			HandleScope scope;

			Local<FunctionTemplate> t = FunctionTemplate::New(New);

			t->Inherit(EventEmitter::constructor_template);
			t->InstanceTemplate()->SetInternalFieldCount(1);

			data_symbol = NODE_PSYMBOL("data");
			play_symbol = NODE_PSYMBOL("play");
			add_symbol = NODE_PSYMBOL("add");
			stop_symbol = NODE_PSYMBOL("stop");
			end_symbol = NODE_PSYMBOL("end");
			
			t->InstanceTemplate()->SetAccessor(String::New("pollRate"), GetPollRate, SetPollRate);

			NODE_SET_PROTOTYPE_METHOD(t, "play", Play);
			NODE_SET_PROTOTYPE_METHOD(t, "playMedia", PlayMedia);
			NODE_SET_PROTOTYPE_METHOD(t, "stop", Stop);
			NODE_SET_PROTOTYPE_METHOD(t, "add", Add);
			NODE_SET_PROTOTYPE_METHOD(t, "pause", Pause);
			NODE_SET_PROTOTYPE_METHOD(t, "next", Next);
			NODE_SET_PROTOTYPE_METHOD(t, "previous", Previous);
			NODE_SET_PROTOTYPE_METHOD(t, "getPlaylist", GetPlaylist);
			NODE_SET_PROTOTYPE_METHOD(t, "parseMedia", ParseMedia);
			NODE_SET_PROTOTYPE_METHOD(t, "insertMedia", InsertMedia);
			NODE_SET_PROTOTYPE_METHOD(t, "removeMedia", RemoveMedia);

			target->Set(String::NewSymbol("NodeVLC"),
				t->GetFunction());
		}

		NodeVLC() : EventEmitter()
		{
		}

		~NodeVLC()
		{
		}

		static Handle<Value> New(const Arguments& args)
		{
			HandleScope scope;
			char const *vlc_argv[] = { "--plugin-path=./vlclibs/vlc/plugins"};
			int vlc_argc = sizeof(vlc_argv) / sizeof(*vlc_argv);
			NodeVLC* vlc = new NodeVLC();
			vlc->inst = libvlc_new(vlc_argc, vlc_argv);
			vlc->media_player = libvlc_media_player_new(vlc->inst);
			vlc->media_list = libvlc_media_list_new(vlc->inst);
			vlc->media_list_player = libvlc_media_list_player_new(vlc->inst);
			libvlc_media_list_player_set_media_player(vlc->media_list_player, vlc->media_player);
			libvlc_media_list_player_set_media_list(vlc->media_list_player, vlc->media_list);
			vlc->poll_rate = 4;
			ev_timer_init(EV_DEFAULT_UC_ &vlc->poll_timer, EIO_poll_timer, 0., 1.0 / vlc->poll_rate);
			vlc->poll_timer.data = vlc;
			vlc->Wrap(args.This());
			return args.This();
		}

		// this.pollRate
		static v8::Handle<Value> GetPollRate(v8::Local<v8::String> property, const v8::AccessorInfo& info)
		{
		// Extract the C++ request object from the JavaScript wrapper.
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(info.Holder());
			return Integer::New(vlc->poll_rate);
		}
		// this.pollRate=
		static void SetPollRate(Local<String> property, Local<Value> value, const AccessorInfo& info)
		{
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(info.Holder());
			
			if (value->IsInt32())
			{
				int newrate = value->ToInteger()->Value();
				vlc->poll_rate = newrate;
			}
		}

		static Handle<Value> ParseMedia(const Arguments& args)
		{
			HandleScope scope;

			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());

			Local<Object> returnObject = Object::New();

			if (args[0]->IsString())
			{
				Local<String> path = Local<String>::New(args[0]->ToString());

				libvlc_media_t * media = libvlc_media_new_path(vlc->inst, *String::Utf8Value(path));
				libvlc_media_parse(media);
				char *title = NULL, *artist = NULL, *genre = NULL, *album = NULL, *tracknum = NULL, *arturl = NULL;
				title = libvlc_media_get_meta(media, libvlc_meta_Title);
				artist = libvlc_media_get_meta(media, libvlc_meta_Artist);
				genre = libvlc_media_get_meta(media, libvlc_meta_Genre);
				album = libvlc_media_get_meta(media, libvlc_meta_Album);
				tracknum = libvlc_media_get_meta(media, libvlc_meta_TrackNumber);
				arturl = libvlc_media_get_meta(media, libvlc_meta_ArtworkURL);
				libvlc_time_t duration = libvlc_media_get_duration(media);

				if (title) { returnObject->Set(String::New("title"), String::New(title)); } 
				if (artist) { returnObject->Set(String::New("artist"), String::New(artist)); }
				if (album) { returnObject->Set(String::New("album"), String::New(album)); }
				if (genre) { returnObject->Set(String::New("genre"), String::New(genre)); }
				if (tracknum) { returnObject->Set(String::New("tracknum"), String::New(tracknum)); }
				if (arturl) { returnObject->Set(String::New("arturl"), String::New(arturl)); }
				if (duration) { returnObject->Set(String::New("duration"), Integer::New(duration)); }

				libvlc_media_release(media);
			}

			return returnObject;
		}

		static Handle<Value> Play(const Arguments& args)
		{
			HandleScope scope;
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());

			libvlc_media_list_player_play(vlc->media_list_player);
			
			if (ev_is_active(&vlc->poll_timer))
			{
				ev_timer_stop(EV_DEFAULT_UC_ &vlc->poll_timer);
			}

			ev_timer_set(EV_DEFAULT_UC_ &vlc->poll_timer, 0., 1.0 / vlc->poll_rate);
			ev_timer_again(&vlc->poll_timer);
			vlc->Emit(play_symbol, 0, NULL);

			return Undefined();
		}

		//Play a track immediately, given a path.
		static Handle<Value> PlayMedia(const Arguments& args)
		{
			HandleScope scope;
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());

			if (args[0]->IsString())
			{
				Local<String> path = args[0]->ToString();
				libvlc_media_t * media = libvlc_media_new_path(vlc->inst, *String::Utf8Value(path));
				
				libvlc_media_player_set_media(vlc->media_player, media);
				libvlc_media_player_play(vlc->media_player);

				if (ev_is_active(&vlc->poll_timer))
				{
					ev_timer_stop(EV_DEFAULT_UC_ &vlc->poll_timer);
				}

				ev_timer_set(EV_DEFAULT_UC_ &vlc->poll_timer, 0., 1.0 / vlc->poll_rate);
				ev_timer_again(&vlc->poll_timer);
				vlc->Emit(play_symbol, 0, NULL);
			}

			return Undefined();
		}

		static Handle<Value> Add(const Arguments& args)
		{
			HandleScope scope;

			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());
			
			if (args[0]->IsArray())
			{
				Local<Array> arr = Array::Cast(*(args[0]->ToObject()));
				int length = arr->Length();

				for (int i=0; i<length; i++)
				{
					Local<String> path = arr->Get(i)->ToString();
					libvlc_media_list_lock(vlc->media_list);
					libvlc_media_t * media = libvlc_media_new_path(vlc->inst, *String::Utf8Value(path));
					libvlc_media_parse(media);
					libvlc_media_list_add_media(vlc->media_list, media);
					libvlc_media_list_unlock(vlc->media_list);			
				}

				Local<Value> outargs[1];
				outargs[0] = arr;
				vlc->Emit(add_symbol, 1, outargs);
			}
			else if (args[0]->IsString())
			{
				Local<String> path = args[0]->ToString();

				libvlc_media_list_lock(vlc->media_list);
				libvlc_media_t * media = libvlc_media_new_path(vlc->inst, *String::Utf8Value(path));
				libvlc_media_parse(media);
				libvlc_media_list_add_media(vlc->media_list, media);
				libvlc_media_list_unlock(vlc->media_list);

				Local<Value> outargs[1];
				outargs[0] = path;
				vlc->Emit(add_symbol, 1, outargs);
			}

			return Undefined();
		}

		static void EIO_poll_timer(EV_P_ ev_timer *w, int revents)
		{
			HandleScope scope;

			NodeVLC *vlc = static_cast<NodeVLC *>(w->data);
			Local<Value> args[1];
			Local<Number> pos = Number::New(libvlc_media_player_get_time(vlc->media_player));
			Local<Object> returnObject = Object::New();
			returnObject->Set(String::New("position"), pos);
			args[0] = returnObject;
			vlc->Emit(data_symbol, 1, args);
		
			if (!libvlc_media_player_is_playing(vlc->media_player) && libvlc_media_player_get_state(vlc->media_player) != libvlc_NothingSpecial)
			{
				ev_timer_stop(EV_A_ w);
				if (libvlc_media_player_get_state(vlc->media_player) == libvlc_Ended)
				{
					vlc->Emit(end_symbol, 0, NULL);
				}
			}
		}

		static Handle<Value> Stop(const Arguments& args)
		{
			HandleScope scope;
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());
			if (ev_is_active(&vlc->poll_timer))
			{
				ev_timer_stop(EV_DEFAULT_UC_ &vlc->poll_timer);
			}
			libvlc_media_list_player_stop(vlc->media_list_player);
			vlc->Emit(stop_symbol, 0, NULL);
			return Undefined();
		}

		static Handle<Value> Pause(const Arguments& args)
		{
			HandleScope scope;
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());
			if (libvlc_media_player_get_state(vlc->media_player) == libvlc_Paused)
			{
				if (ev_is_active(&vlc->poll_timer))
				{
					ev_timer_stop(EV_DEFAULT_UC_ &vlc->poll_timer);
				}

				ev_timer_set(EV_DEFAULT_UC_ &vlc->poll_timer, 0., 1.0 / vlc->poll_rate);
				ev_timer_again(&vlc->poll_timer);
				vlc->Emit(play_symbol, 0, NULL);
			}
			libvlc_media_list_player_pause(vlc->media_list_player);
			return Undefined();
		}

		static Handle<Value> Next(const Arguments& args)
		{
			HandleScope scope;
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());
			libvlc_media_list_player_next(vlc->media_list_player);
			return Undefined();
		}

		static Handle<Value> Previous(const Arguments& args)
		{
			HandleScope scope;
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());
			libvlc_media_list_player_previous(vlc->media_list_player);
			return Undefined();
		}

		static Handle<Value> InsertMedia(const Arguments& args)
		{
			HandleScope scope;
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());
			if (args[1]->IsNumber() && args[0]->IsString())
			{
				libvlc_media_t *media = libvlc_media_new_path(vlc->inst, *String::Utf8Value(args[0]->ToString()));
				libvlc_media_parse(media);
				libvlc_media_list_lock(vlc->media_list);
				libvlc_media_list_insert_media(vlc->media_list, media, args[1]->ToInt32()->Value());
				libvlc_media_list_unlock(vlc->media_list);
			}
			return Undefined();
		}

		static Handle<Value> RemoveMedia(const Arguments& args)
		{
			HandleScope scope;
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());
			if (args[0]->IsNumber())
			{
				libvlc_media_list_lock(vlc->media_list);
				libvlc_media_list_remove_index(vlc->media_list, args[0]->ToInt32()->Value());
				libvlc_media_list_unlock(vlc->media_list);
			}
			return Undefined();
		}

		static Handle<Value> GetPlaylist(const Arguments& args)
		{
			HandleScope scope;
			NodeVLC* vlc = ObjectWrap::Unwrap<NodeVLC>(args.This());
			int count = libvlc_media_list_count(vlc->media_list);

			Local<Array> playlist = Array::New(count);
			for (int i=0; i < count; i++)
			{
				Local<Object> mediainfo = Object::New();
				libvlc_media_t * media = libvlc_media_list_item_at_index(vlc->media_list, i);
				char * mrl = libvlc_media_get_mrl(media);
				char * artist = libvlc_media_get_meta(media, libvlc_meta_Artist);
				char * title = libvlc_media_get_meta(media, libvlc_meta_Title);
				char * album = libvlc_media_get_meta(media, libvlc_meta_Album);
				char * genre = libvlc_media_get_meta(media, libvlc_meta_Genre);
				if (mrl) { mediainfo->Set(String::New("mrl"), String::New(mrl)); }
				if (artist) { mediainfo->Set(String::New("artist"), String::New(artist)); }
				if (title) { mediainfo->Set(String::New("title"), String::New(title)); }
				if (album) { mediainfo->Set(String::New("album"), String::New(album)); }		
				if (genre) { mediainfo->Set(String::New("genre"), String::New(genre)); }
				
				playlist->Set(i, mediainfo);
			}

			return playlist;
		}

};

extern "C" {
	static void init (Handle<Object> target)
	{
		NodeVLC::Init(target);
	}

	NODE_MODULE(nodevlc, init);
}
